# Copyright 2020 Ecosoft Co., Ltd (https://ecosoft.co.th/)
# License AGPL-3.0 or later (https://www.gnu.org/licenses/agpl.html)
from odoo import _, api, fields, models
from odoo.exceptions import ValidationError


class AccountWithholdingTax(models.Model):
    _name = "account.withholding.tax"
    _description = "Account Withholding Tax"

    name = fields.Char(required=True)
    account_id = fields.Many2one(
        comodel_name="account.account",
        string="Withholding Tax Account",
        domain=[("wht_account", "=", True)],
        required=True,
        ondelete="restrict",
    )
    amount = fields.Float(
        string="Percent",
    )
    is_pit = fields.Boolean(
        string="Personal Income Tax",
        help="As PIT, the calculation of withholding amount is based on pit.rate",
    )
    pit_id = fields.Many2one(
        comodel_name="personal.income.tax",
        string="PIT Rate",
        compute="_compute_pit_id",
        help="Latest PIT Rates used to calcuate withholiding amount",
    )
    _sql_constraints = [
        ("name_unique", "UNIQUE(name)", "Name must be unique!"),
    ]

    @api.constrains("is_pit")
    def _check_is_pit(self):
        pits = self.search_count([("is_pit", "=", True)])
        if pits > 1:
            raise ValidationError(_("Only 1 personal income tax allowed!"))

    @api.constrains("account_id")
    def _check_account_id(self):
        for rec in self:
            if rec.account_id and not rec.account_id.wht_account:
                raise ValidationError(_("Selected account is not for withholding tax"))

    @api.depends("is_pit")
    def _compute_pit_id(self):
        pit_date = self.env.context.get("pit_date") or fields.Date.context_today(self)
        pit = self.env["personal.income.tax"].search(
            [("effective_date", "<=", pit_date)], order="effective_date desc", limit=1
        )
        self.update({"pit_id": pit.id})
